/***************************************************************
                     _                _
                __ _| |_ __ _   _ __ (_) ___
               / _` | __/ _` | | '_ \| |/ _ \
              | (_| | || (_| | | |_) | | (_) |
               \__,_|\__\__,_| | .__/|_|\___/
                               |_|  pata hard disk support
 ***************************************************************
 It's intended that this be kept simple in design. It's using
 the inefficient PIO access, only supports a single disk, and
 does not attempt to allow sharing the disk with other operating
 systems.
 **************************************************************/

#define ATA_READ           0x20
#define ATA_WRITE          0x30
#define ATA_FLUSH          0xE7
#define ATA_PRIMARY        0x1F0
#define ATA_DATA           0x1F0
#define ATA_ERROR          0x1F1
#define ATA_FEATURES       0x1F1
#define ATA_SECTOR_COUNT   0x1F2
#define ATA_SECTOR_NUMBER  0x1F3
#define ATA_CYLINDER_LOW   0x1F4
#define ATA_CYLINDER_HIGH  0x1F5
#define ATA_DRIVE          0x1F6
#define ATA_HEAD           0x1F6
#define ATA_STATUS         0x1F7
#define ATA_COMMAND        0x1F7

int sx, sa;

void ata_wait_bsy() { while (inb(ATA_STATUS) & 0x80); }
void ata_wait_drq() { while (!(inb(ATA_STATUS) & 0x40)); }

void ata_request(int n, unsigned char req) {
  ata_wait_bsy();
  outb(ATA_HEAD,          0xE0 | ((n >> 24) &  0x0F));
  outb(ATA_ERROR,         0x00);
  outb(ATA_SECTOR_COUNT,  0x08);
  outb(ATA_SECTOR_NUMBER, (unsigned char)n);
  outb(ATA_CYLINDER_LOW,  (unsigned char)(n >> 8));
  outb(ATA_CYLINDER_HIGH, (unsigned char)(n >> 16));
  outb(ATA_COMMAND, req);
}

void ata_read(int n, char *buffer) {
  ata_request(n, ATA_READ);
  for (sx = 0; sx < 8; sx++) {
    ata_wait_bsy();
    ata_wait_drq();
    for (sa = 0; sa < 512; sa += 2) {
      buffer[(sx * 512) + sa] = inw(ATA_PRIMARY);
    }
  }
}

void ata_write(int n, char *buffer) {
  ata_request(n, ATA_WRITE);
  for (sx = 0; sx < 8; sx++) {
    ata_wait_bsy();
    ata_wait_drq();
    for (sa = 0; sa < 512; sa += 2) {
      outw(ATA_PRIMARY, buffer[(sx * 512) + sa]);
    }
  }
}
