/***************************************************************
           _                      _             _
          | |_ ___ _ __ _ __ ___ (_)_ __   __ _| |
          | __/ _ \ '__| '_ ` _ \| | '_ \ / _` | |
          | ||  __/ |  | | | | | | | | | | (_| | |
           \__\___|_|  |_| |_| |_|_|_| |_|\__,_|_|
          a limited subset of a DEC-style terminal
 ***************************************************************
 This builds on top of the barebones VGA text mode device. It
 implements a subset of the DEC VT series escape sequences. The
 support is limited to basic foreground/background colors,
 clearing the display, and basic cursor movement.
 **************************************************************/

I acc[32], acc_i = 0, ground = 0, on = 0;

I acc_pop() {
  I k = acc_i;
  if (acc_i > 0) {
    acc_i -= 1;
    if (acc_i == -1) {
      acc_i = 0;
    }
  }
  return acc[k];
}

V acc_reset() {
  for (I x = 0; x < 32; x++) {
    acc[x] = 0;
  }
  ground = 0;
  on = 0;
}

V gt_up(I n)       { ty -= 1; if (ty < 0) ty = 0;   cursor(); }
V gt_down(I n)     { ty += 1; if (ty > 24) ty = 24; cursor(); }
V gt_left(I n)     { tx -= 1; if (tx < 0) tx = 0;   cursor(); }
V gt_right(I n)    { tx += 1; if (tx > 79) ty = 79; cursor(); }
V gt_move_cursor() {
  if (acc_i >= 1) {
    I a = acc_pop();
    I b = acc_pop();
    ty = b - 1; if (ty < 0) ty = 0; if (ty > 24) ty = 24;
    tx = a - 1; if (tx < 0) tx = 0; if (tx > 79) tx = 79;
    cursor();
    acc_reset();
    return;
  }
  /* home */
  acc_reset(); cursor();
}

V gt_set_fg(I c)   { c = c - 30; attr = attr & 0xF0;
                     attr = attr + c; acc_reset(); }
V gt_set_bg(I c)   { c = c - 40; attr = attr & 0x0F;
                     attr = attr + (c << 4); acc_reset(); }
V gt_reset()       { acc_reset(); attr = 0x06; }
V gt_next()        { acc_i += 1; acc[acc_i] = 0; }
V gt_clear()       { acc_reset(); vga_init(); }
V gt_set_attr()    {
  if (acc_pop() == 0) { gt_reset(); return; }
  ground == 0 ? gt_set_fg(acc_pop()) : gt_set_bg(acc_pop());
  acc_reset();
}

V term_putc(I c) {
  if (c == 27) {
    on = -2;
    return;
  }
  switch (on) {
    case 0:
      display(c);
      return;
    case -2:
      if (c == '[') {
        on = -1;
        return;
      }
      on = 0;
      display(c);
      return;
  }
  if (c >= '0' && c <= '9') {
    acc[acc_i] *= 10;
    acc[acc_i] += c - 48;
    if (c == '3') {
      ground = 0;
    }
    if (c == '4') {
      ground = 1;
    }
  } else {
    switch (c) {
      case 'A': gt_up(acc_pop());    break;
      case 'B': gt_down(acc_pop());  break;
      case 'C': gt_left(acc_pop());  break;
      case 'D': gt_right(acc_pop()); break;
      case 'm': gt_set_attr();       break;
      case ';': gt_next();           break;
      case 'H': gt_move_cursor();    break;
      case 'J': gt_clear();          break;
      default: on = 0;
               display(c);
    }
  }
  cursor();
}
